/*
 * BUFFER - Zero copy buffers (non-wrapping rings)
 *
 * Theory is as follows. You get direct access to write and read pointers,
 * macros to advance them (put and get), and macros to get the amount of bytes
 * available for reading and writing (maxput and maxget).
 *
 * Whenever the read pointer catches up with the write pointer, both are
 * reset to the start of the buffer.
 *
 * It is required that you don't write more data at the write pointer than
 * there's room as given by the 'maxput' macro. No bounds checking is done
 * by the 'put' macro, because the damage would already be done anyway. 
 *
 * The same goes for reading; you need to call the 'maxget' macro to know how
 * many bytes you may access at the read pointer and how far you may advance 
 * it using the 'get' macro.
 *
 * If you don't trust yourself to do that faithfully, you can always add a
 * wrapper above this that offers a classic check, copy, advance API, but
 * of course you lose the zero copy property that way.
 *
 * If you use multithreading, /you/ are responsible for making sure no one else
 * writes to your buffer between looking at buf_maxput and calling buf_put, and
 * that no one else reads from it between looking at buf_maxget and calling
 * buf_get. The operations themselves are thread safe.
 *
 * Author:
 * Emile van Bergen, emile@e-advies.nl
 *
 * License:
 * Permission to redistribute an original or modified version of this program
 * in source, intermediate or object code form is hereby granted exclusively
 * under the terms of the GNU General Public License, version 2. Please see the
 * file COPYING for details, or refer to http://www.gnu.org/copyleft/gpl.html.
 *
 * History:
 * 2006/02/15 - EvB - Created
 */


/* 
 * INCLUDES & DEFINES
 */


#include <evblib/str/str.h>


#define buf_capacity(B)   ((B)->bu.l)
#define buf_maxget(B)	  ((B)->w - (B)->r)
#define buf_maxput(B)	  ((B)->bu.p + (B)->bu.l - (B)->w)
#define buf_empty(B)	  ((B)->r = (B)->w = (B)->bu.p)

#define buf_get(B,len) 	  (void)(((B)->r += (len)) == (B)->w && buf_empty(B))
#define buf_put(B,len)	  (void)((B)->w += (len))


/* 
 * TYPES
 */


typedef struct buf {
    STR_T bu;
    char *r, *w;
} BUF_T;


/*
 * PROTOTYPES
 */


BUF_T *buf_new(size_t size);			/* malloc + init */
int buf_init(BUF_T *b, size_t size);		/* 0 if succes, < 0 if error */
void buf_done(BUF_T *b);
void buf_del(BUF_T *b);				/* done + free */

