
#include <evblib/sockets/sockets.h>

#include <evblib/str/str.h>
#include <evblib/misc/misc.h>
#include <evblib/sysdefs/byteorder.h>

#include <sys/utsname.h>
#include <netdb.h>

#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>


/* returns listening socket on success, -errno on error. */

int listensock_init( int addrfamily, char *addrp, size_t addrl,
				   int socktype, int protocol, int port ) 
{
	struct sockaddr_in *sin;
	size_t sinlen;
	int retfd, e;

	/* allocate socket address structure; length is at least that of 
       sockaddr_in and sufficient for addrl bytes to fit starting at
	   the sin_addr member. */

	sin = 0;
	sinlen = sizeof( struct sockaddr_in );
	if ( (char *)&sin->sin_addr + addrl > (char *)sin + sinlen )
		sinlen = (char *)&sin->sin_addr - (char *)sin + addrl;
	sin = (struct sockaddr_in *)malloc( sinlen );
	if ( sin == 0 ) { return -errno; }

    /* create the socket */

	retfd = socket( addrfamily, socktype, protocol );
	if ( retfd == -1 ) { e = errno; goto err_freesin; }

	/* children don't need the listening socket; set close on exec; also
	   set SO_REUSEADDR */

#ifndef WINDOWS
	fcntl( retfd, F_SETFD, 1 );
#endif
	e = 1;
	setsockopt( retfd, SOL_SOCKET, SO_REUSEADDR, (char *)&e, sizeof(e) );

	/* initialize the address structure */

	memset( sin, 0, sinlen );
	sin->sin_family = addrfamily;
	sin->sin_port = net16( port );
	memcpy( &sin->sin_addr, addrp, addrl );

	/* bind the socket to the given address, mark it as listening and return */

    if ( bind( retfd, (struct sockaddr *)sin, sinlen ) != -1 &&
	     listen( retfd, SOMAXCONN ) != -1 ) 
	{ 
		free( sin );
		return retfd;
	}

	/* we only get here if the bind or listen fails */

	e = errno;
	close( retfd );

err_freesin:
	free( sin ); 
	return -e;
}

int listensock_init_ipv4( uint32_t ipv4, int socktype, int protocol, int port )
{
	uint32_t ipv4be = net32( ipv4 );
	return listensock_init( AF_INET, (char *)&ipv4be, sizeof( ipv4be ),
							socktype, protocol, port );
}

/* returns connected socket on success, -errno on error. */

int accept_ipv4( int lsfd, uint32_t *fromipv4, int *fromport )
{
	struct sockaddr_in sin;
	int retfd; 
	socklen_t sinlen;

	sinlen = sizeof( sin );
	retfd = accept( lsfd, (struct sockaddr *)&sin, &sinlen );
	if ( retfd == -1 ) return -errno;

	if ( fromipv4 ) *fromipv4 = getnet32( &sin.sin_addr );
	if ( fromport ) *fromport = net16( sin.sin_port );
	return retfd;
}

uint32_t ownipv4addr()
{
	struct utsname uts;
	uint32_t ret;
	int err;

	if ( uname( &uts ) == -1 ) return 0;
	err = 0;
	ret = getipv4addrbyname( uts.nodename, strlen( uts.nodename ), &err );
	if ( err ) return 0;
	return ret;
}

int socket_init( int addrfamily, int socktype, int protocol )
{
	int ret;

	ret = socket( addrfamily, socktype, protocol );
	if ( ret == -1 ) ret = -errno;
	return ret;
}

int socket_connect( int sockfd, int addrfamily, char *addrp, size_t addrl,
					int destport )
{
	struct sockaddr_in *sin;
	size_t sinlen;
	int ret;

	/* allocate socket address structure; length is at least that of 
       sockaddr_in and sufficient for addrl bytes to fit starting at
	   the sin_addr member. */

	sin = 0;
	sinlen = sizeof( struct sockaddr_in );
	if ( (char *)&sin->sin_addr + addrl > (char *)sin + sinlen )
		sinlen = (char *)&sin->sin_addr - (char *)sin + addrl;
	sin = (struct sockaddr_in *)malloc( sinlen );
	if ( sin == 0 ) { return -errno; }

	/* initialize the address structure */

	memset( sin, 0, sinlen );
	sin->sin_family = addrfamily;
	sin->sin_port = net16( destport );
	memcpy( &sin->sin_addr, addrp, addrl );

	ret = connect( sockfd, (struct sockaddr *)sin, sinlen );
	free( sin );
	if ( ret == -1 ) ret = -errno;
	return ret;
}

int socket_connect_ipv4( int sockfd, uint32_t ipv4addr, int destport )
{
	uint32_t ipv4be = net32( ipv4addr );
	return socket_connect( sockfd, AF_INET, (char *)&ipv4be, sizeof( ipv4be ),
						   destport );
}

int setblockingmode( int fd, int blocking )
{
	if ( fcntl( fd, F_SETFL, blocking ? 0 : O_NONBLOCK ) == -1 ) 
		return -errno;
	return 0;
}

STR_T ipv4tostr( uint32_t ipv4 )
{
	STR_T ret;

	/* we use snprintf for now */

	ret.l = 24;
	ret.p = (char *)malloc( ret.l );
	if ( ret.p == 0 ) { ret.l = 0; return ret; }

	snprintf( ret.p, ret.l, "%u.%u.%u.%u", ipv4 >> 24, 
										   ( ipv4 >> 16 ) & 0xff,
										   ( ipv4 >> 8 ) & 0xff,
										   ipv4 & 0xff );
	ret.l = strlen( ret.p );
	return ret;
}

uint32_t strtoipv4( char *addrstr, size_t addrstrl, 
				    int *ret_error )
{
	struct in_addr addr;
	uint32_t ret;
	char *name;
	int err;

	ret = err = 0;
	name = str_dupt( addrstr, addrstrl );
	if ( ! name ) { err = -errno; goto done; }

	if ( inet_aton( name, &addr ) ) ret = net32( addr.s_addr );
	else err = -EADDRNOTAVAIL;
	free( name );

done:
	if ( ret_error ) *ret_error = err;
	return ret;
}

uint32_t getipv4addrbyname( char *hostname, size_t hostnamel, int *ret_error )
{
	struct hostent *he;
	char *name;
	int err;
	uint32_t ret;

	/* first try dotted-decimal conversion */

	err = 0;
	ret = strtoipv4( hostname, hostnamel, &err );
	if ( ! err ) goto done;

	/* no luck, try a lookup */

	errno = 0;
	name = str_dupt( hostname, hostnamel );
	if ( ! name ) { err = -errno; goto done; }

	errno = 0;
	he = gethostbyname( name );
	free( name );
	if ( ! he ) { err = -errno; goto done; }
	if ( he->h_addrtype != AF_INET ||
		 he->h_length != 4 ||
		 he->h_addr == 0 ) { err = -EDOM; goto done; }
	ret = getnet32( he->h_addr );
	err = 0;

done:
	if ( ret_error ) *ret_error = err;
	return ret;
}

	
/*
 * vim:ts=4:sw=4
 */

