/*
 * STREAM - Functions to support efficient I/O streams based on fastrings
 *
 * Author:
 * Emile van Bergen, emile@evbergen.xs4all.nl
 *
 * Permission to redistribute an original or modified version of this program
 * in source, intermediate or object code form is hereby granted exclusively
 * under the terms of the GNU General Public License, version 2. Please see the
 * file COPYING for details, or refer to http://www.gnu.org/copyleft/gpl.html.
 *
 * History:
 * 2004/01/13 - EvB - Created
 */

char stream_id[] = "STREAM - Copyright (C) 2004 Emile van Bergen.";


/*
 * INCLUDES & DEFINES
 */


#include <string.h>		/* for memchr */
#include <stdlib.h>		/* for malloc */
#include <errno.h>		/* for errno */

#include <evblib/stream/stream.h>
#include <evblib/misc/pagesize.h>
#include <evblib/misc/misc.h>	/* for safe_read */


/* 
 * FUNCTIONS
 */


STRM_T *strm_new(int fd, ssize_t bufsize)
{
    STRM_T *ret;

    ret = (STRM_T *)malloc(sizeof(STRM_T)); 
    if (ret) {
	if (strm_init(ret, fd, bufsize) >= 0) 
	    return ret;
        free(ret);
    }
    return 0;
}

int strm_init(STRM_T *strm, int fd, ssize_t bufsize)
{
    size_t pagesize = GETPAGESIZE;
 
    memset(strm, 0, sizeof(*strm));
    strm->bufsize = ((bufsize + pagesize - 1) / pagesize) * pagesize;
    strm->fd = fd;
    return ring_init(&strm->r, strm->bufsize * 2);
}

void strm_done(STRM_T *strm)
{
    ring_done(&strm->r);
}

void strm_del(STRM_T *strm)
{
    strm_done(strm);
    free(strm);
}


STR_T strm_getline(STRM_T *strm)
{
    ssize_t max, len, rd, rl;
    char *end, *rp;
    
    for(;;) {
	max = ring_maxget(&strm->r);
	end = memchr(strm->r.r, '\n', max);
	if (end) { rl = end - strm->r.r + 1; break; }

	if (strm->eof) { rl = max; break; }
        rd = ring_maxput(&strm->r); 
	if (!rd) { rl = max; break; }	   /* line cut in two */
	if (rd > strm->bufsize) rd = strm->bufsize;
	len = safe_read(strm->fd, strm->r.w, rd);
	ring_put(&strm->r, len);
	if (!len) strm->eof++;
    }
    strm->aux++;
    rp = strm->r.r; 
    ring_get(&strm->r, rl);
    return (STR_T){rp, rl};
}


/* Ensures len bytes minimum available in buffer; if len = 0 try to get as much
   as possible; if len < 0 return what we have now, without pulling; otherwise 
   we return the amount we have up to requested amount, after pulling. */

STR_T strm_ensure(STRM_T *strm, ssize_t len)
{
    ssize_t max, hasread, toread;
    STR_T ret;

    if (len == 0) len = strm->bufsize;
    for(;;) {
	max = ring_maxget(&strm->r); 
	if (max >= len) { ret.l = len >= 0 ? len : max; break; }

	if (strm->eof)  { ret.l = max; break; }
	toread = ring_maxput(&strm->r); 
	if (!toread) { ret.l = max; break; }
	hasread = safe_read(strm->fd, strm->r.w, toread);
	if (hasread) { ring_put(&strm->r, hasread); continue; }
	strm->eof++;
    }
    ret.p = strm->r.r;
    return ret;
}

/* Adds as much data to the ring as we can using a single read, as limited by
   the space available and the amount of data pending in the kernel. Returns 
   all data we have in the ring. If ret_status is nonzero, the integer at the 
   address given by it is set to one of the following values:

	status < 0:	status equals -errno; test against EAGAIN if you're 
			using non-blocking I/O and against EINTR if you're 
			using blocking I/O
	status = 0:	end of file on file descriptor
	status > 0:	amount of data added to the ring 

   If this function is called without any buffer space being available, no
   read is performed and -EMSGSIZE isreturned in *ret_status. 
   After eof has been seen and reported, every subsequent strm_read returns 
   EOF in ret_status without performing any read, until strm->eof is reset.
*/

STR_T strm_read(STRM_T *strm, int *ret_status)
{
    int maxread, res;
    STR_T ret;

    if (strm->eof) { res = 0; goto done; }

    maxread = ring_maxput(&strm->r); 
    if (maxread == 0) { res = -EMSGSIZE; goto done; }

    res = read(strm->fd, strm->r.w, maxread);
    if (res > 0) ring_put(&strm->r, res);
    else if (res < 0) res = -errno; 
    else strm->eof++;

done:
    ret.p = strm->r.r;
    ret.l = ring_maxget(&strm->r);
    if (ret_status) *ret_status = res;
    return ret;
}

/*
 * vim:softtabstop=4:sw=4
 */

